/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

extern class ModelFB model;
extern int vMAXnbITERATIONS;
extern int vMINnbITERATIONS;
extern int maxlen;

extern float vMINdifftotLogLik;
extern bool alloutput, storeseqs, bdsDNA, bdsDNAjointly, bUseExtP;

#include "BWmain.h"

extern bool verbose, optCmatrix;	
extern  int  Code[], Decode[];
extern bool FBDdetailsC;
extern double thresholds[6];
extern int    thresholdNb;

extern  int   seqNb, seqLen;
extern  int   seq[kMaxSeqLen+1];
extern  bool  completedFile;
extern  char  SeqName[kMaxSeqName+2];
extern  FILE *cFpWriteDom; 
extern  FILE *cFpWritePP, *cFpWritePPD,*cFpWriteCP; 
extern  FILE *cFpReadSeq;

int iterationNb;
long double totLogLik, origtotLogLik, prevtotLogLik, difftotLogLik;
FILE *cFpWriteBW;

bool CheckStop(void);
void MonitorLikChange(void);


// ---------------------------------------------------------------------------
//		 BWmain 				
// ---------------------------------------------------------------------------
void BWmain(const char *ModelFile, char *seqFile, const int nbseqs, const int maxlen )
{
bool stoppingCondition;
int i, j, totseqnb=0;
int **seqs=NULL; 
int *seqsmem;
int *seqsLen=NULL;

FBDdetailsC	= true; // suppress output in Baum-Welch

if (storeseqs)// keep seqs in memory
	{
	if ((nbseqs>0) && (maxlen>0))
		{
		seqsmem=(int *) malloc(nbseqs*(maxlen+1)*sizeof(int));
		if (seqsmem==NULL) {cerr<<"\nFailed in malloc seqsmem in BWmain"<<endl;exit(0);}
		seqs=(int **) malloc(nbseqs*sizeof(int *));
		if (seqs==NULL) {cerr<<"\nFailed in malloc seqsin BWmain"<<endl;exit(0);}
		for (i=0; i<nbseqs; i++) seqs[i]=seqsmem+i*(maxlen+1);
		for (i=0; i<nbseqs; i++)
			{
			for (j=0; j<=maxlen; j++)  seqs[i][j]=0;
			 }

		seqsLen=(int *) malloc(nbseqs*sizeof(int));
		if (seqsmem==NULL) {cerr<<"\nFailed in malloc seqsLen in BWmain"<<endl;exit(0);}
		}
	else {cerr<<"\nThe nb of sequences and their maximal length are required with option -k"<<endl;exit(0);}
	}

iterationNb = 0;stoppingCondition = false;

if  ((cFpWriteBW = fopen( "InitialModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file InitialModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#BEFORE BAUM-WELCH, INITIAL MODEL \n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

if (alloutput)
{
if  ((cFpWriteBW = fopen( "UpdateModels.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file UpdateModels for writing"<<endl; exit(1);}
}
	 
if  ((cFpReadSeq = fopen( seqFile, "r" ))  == NULL )
		{cerr << "\n could not open the file for the sequences with address  " << 
					seqFile <<endl;  exit(1);}
	 
while (stoppingCondition == false )
	{
	model.SetUpScaledParameters();
	model.SetUpConnections();
	model.InitialiseBWIterStep();	
	seqNb = 0;totLogLik=0;
	completedFile = false;iterationNb++;

	if ((storeseqs) && (iterationNb>1))    //seqs from memory
		{
		while (seqNb < totseqnb)
			{
			seqNb ++;
			seqLen = seqsLen[seqNb];
			if (seqLen > 0)
				{ 
				  if (bUseExtP)
	  				{
        			 if (seqLen > kMaxExtPLen)
			{cerr<<"\n Sequence Length excessive for use of external prob matrix, limit= "<<kMaxExtPLen<<endl;}
					 else 
		    			{
						model.FBalgorithmExtP(seqNb, seqLen, seqs[seqNb]);
						if (model.bFBseqProbPositive) totLogLik+= model.logFBseqProb;
						model.ComputeSeqBW(seqNb, seqLen, seqs[seqNb]); 
						}
					seqNb = totseqnb;// only one sequence with bUseExtP
					}
				  else 
					{
					model.FBalgorithm(seqNb, seqLen, seqs[seqNb]);
					if (model.bFBseqProbPositive) totLogLik+= model.logFBseqProb;
					model.ComputeSeqBW(seqNb, seqLen, seqs[seqNb]); 
					}
				}	
			}
		}
		
	else
		{
		rewind(cFpReadSeq);
		while (completedFile == false)
			{
			seqNb ++;
			if ( !(ReadSeq (seqNb, seqLen, seq, completedFile, SeqName)) )
			  {
			    if (seqLen > 0)
			      { WriteWarning(seqNb,1);}	
			    else
			      {seqNb--;}
			  } 
			if (seqLen > 0)
				{ 
				  if (bUseExtP)
	  				{
        			 if (seqLen > kMaxExtPLen)
			{cerr<<"\n Sequence Length excessive for use of external prob matrix, limit= "<<kMaxExtPLen<<endl;}
					 else 
		    			{
						model.FBalgorithmExtP(seqNb, seqLen, seq);
						if (model.bFBseqProbPositive) totLogLik+= model.logFBseqProb;
						model.ComputeSeqBW(seqNb, seqLen, seq); 
						}
					completedFile = true;// only one sequence with bUseExtP
					}

				  else 
					{
					model.FBalgorithm(seqNb, seqLen, seq);
					if (model.bFBseqProbPositive) totLogLik+= model.logFBseqProb;
					model.ComputeSeqBW(seqNb, seqLen, seq); 
					}
				}	
			if ((storeseqs) && (iterationNb==1))
				{
				for (j=1; j<=seqLen; j++)  seqs[seqNb][j]=seq[j];
				seqsLen[seqNb] = seqLen;
				}
			}
		totseqnb = 	seqNb;
		}
	
	if (iterationNb==1) {prevtotLogLik=origtotLogLik=totLogLik;}
	model.ParameterIterationBW();
	if (alloutput) 
		{
		fprintf( cFpWriteBW,"\n#AFTER ITERATION NB=%d\n", iterationNb);
		model.WriteModelFB(cFpWriteBW);
		}
	MonitorLikChange();
	stoppingCondition = CheckStop();
	}
	
if (alloutput)
{ 
fprintf(cFpWriteBW,"\n#AFTER LAST ITERATION\n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);
}

if  ((cFpWriteBW = fopen( "FinalModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file FinalModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#AFTER BAUM-WELCH, FINAL MODEL \n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

fclose(cFpReadSeq);
}
/* BWmain BAUM-WELCH
************************************/


// ---------------------------------------------------------------------------
//		 BWmaindsDNA 				
// ---------------------------------------------------------------------------
void BWmaindsDNA(const char *ModelFile, char *seqFile, const int nbseqs, const int maxlen )
{
bool stoppingCondition;
int totseqnb;
long double firststrandloglik, secondstrandloglik, diff, secondstrandweight, P=0, Q;

if (verbose) cerr << "BWmaindsDNA" << endl;
FBDdetailsC	= true; 

iterationNb = 0;stoppingCondition = false;
if  ((cFpWriteBW = fopen( "InitialModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file InitialModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#BEFORE BAUM-WELCH, INITIAL MODEL \n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

if (alloutput)
{
if  ((cFpWriteBW = fopen( "UpdateModels.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file UpdateModels for writing"<<endl; exit(1);}
}
	
if  ((cFpReadSeq = fopen( seqFile, "r" ))  == NULL )
		{cerr << "\n could not open the file for the sequences with address  " << 
					seqFile <<endl;  exit(1);}
	 
while (stoppingCondition == false )
	{
	model.SetUpScaledParameters();
	model.SetUpConnections();
	model.InitialiseBWIterStep();	
	rewind(cFpReadSeq);
	seqNb = 0;totLogLik=0;
	completedFile = false;iterationNb++;

	rewind(cFpReadSeq);
	while (completedFile == false)
		{
		seqNb ++;
		if ( !(ReadSeq (seqNb, seqLen, seq, completedFile, SeqName)) )
			{
			if (seqLen > 0)	{ WriteWarning(seqNb,1);}	
			else	{seqNb--;}
			} 
		if (seqLen > 0)
			{ 
			model.FBalgorithm(seqNb, seqLen, seq);
			if (bdsDNA) {if (model.bFBseqProbPositive) totLogLik+= model.logFBseqProb;}
			//AS INDEPENDENT SEQUENCES 
			else {if (model.bFBseqProbPositive) {P = exp(model.logFBseqProb);}
					else P=0; }
			firststrandloglik = model.logFBseqProb;
			model.ComputeSeqBWdsDNA1(seqNb, seqLen, seq);
			model.rcseq(seqLen, seq);
			model.FBalgorithm(seqNb, seqLen, model.SecondStrand);
			secondstrandloglik = model.logFBseqProb;
			diff = firststrandloglik - secondstrandloglik; 
			if  (bdsDNAjointly) {secondstrandweight = 1.0 / (1.0 + ((long double) exp ( (double) diff)));}
			else {secondstrandweight = 0.5 ; }
			model.ComputeSeqBWdsDNA2(seqNb, seqLen, model.SecondStrand, secondstrandweight);
			if (bdsDNA) {if (model.bFBseqProbPositive) totLogLik+= model.logFBseqProb;}//AS INDEPENDENT SEQUENCES 
			else 
				{
				if (model.bFBseqProbPositive) {Q = exp(model.logFBseqProb);}
				else Q=0; 
				P += Q;
				if (P>0) {totLogLik+= log(P);}
				 }
			}	
		}
		totseqnb = 	seqNb;
	
	if (iterationNb==1) {prevtotLogLik=origtotLogLik=totLogLik;}
	model.ParameterIterationBW();

	if (alloutput) 
		{
		fprintf( cFpWriteBW,"\n#AFTER ITERATION NB=%d\n", iterationNb);
		model.WriteModelFB(cFpWriteBW);
		}
	MonitorLikChange();
	stoppingCondition = CheckStop();
	}
	
if (alloutput)
{ 
fprintf(cFpWriteBW,"\n#AFTER LAST ITERATION\n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);
}

if  ((cFpWriteBW = fopen( "FinalModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file FinalModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#AFTER BAUM-WELCH, FINAL MODEL \n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

fclose(cFpReadSeq);
}
/* BAUM-WELCH  BWmaindsDNA
************************************/
 
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
//		CheckStop				
// ---------------------------------------------------------------------------
bool   CheckStop(void) 
{
bool  decision;

decision = false;
if (iterationNb >= vMAXnbITERATIONS)  decision = true;
if ((iterationNb >= vMINnbITERATIONS) && ( fabs(difftotLogLik) < vMINdifftotLogLik))  decision = true;
if (iterationNb < 2) decision = false;	

if (verbose) cerr << "End CheckStop" << endl;	
return(decision);
}
/* CheckStop
************************************/


// ---------------------------------------------------------------------------
//		MonitorLikChange	> 				
// ---------------------------------------------------------------------------
void MonitorLikChange(void)
{
difftotLogLik=totLogLik-prevtotLogLik;
cout<<"\n logLik of the learning set of sequences  " << totLogLik << "  at evaluation " <<  iterationNb;
cout<<"\n difftotLogLik of sequence set=  " << difftotLogLik;
cout<<"\n origtotLogLik of sequence set=  " << origtotLogLik<< endl;

if (alloutput) 
	{fprintf(cFpWriteBW,  "\nORIG logLik=%g \nPREV logLik=%g \t PRES logLik=%g \t DIFF logLik=%g \t   "  ,
		(double) origtotLogLik, (double) prevtotLogLik,(double) totLogLik, (double) difftotLogLik);
	} 
	
prevtotLogLik=totLogLik;
}
/* MonitorLikChange
************************************/
// ---------------------------------------------------------------------------

