/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

extern class ModelFB model;
#include "FBmain.h"

extern double CutoffForWriting;
extern bool   verbose, optCmatrix, bstateweights, bUseExtP;	
extern int    Code[], Decode[];
extern bool   FBDdetailsC;
extern double thresholds[6];
extern int    thresholdNb;

extern int   seqNb, seqLen;
extern int   seq[kMaxSeqLen+1];
extern bool  completedFile;
extern char  SeqName[kMaxSeqName+2];
extern FILE *cFpReadSeq;
extern FILE *cFpWriteCP; 

extern bool   bdsDNA;

FILE *cFpWriteFB;

void  WriteProbProfile(int len, const int Seq[]);

// ---------------------------------------------------------------------------
//		 FBmain 	forward-backward posterior decoding
// ---------------------------------------------------------------------------
void FBmain(const char *ModelFile, char *seqFile )
{
  char *tok;

  if (verbose)  cerr  <<  "\nStarting FBmain" <<  endl;

  model.SetUpConnections();
  
  if (bstateweights)  model.SetUpScaledParametersW();
  else model.SetUpScaledParameters();

  if  ((cFpReadSeq = fopen( seqFile, "r" ))  == NULL )
    { cerr << "\n could not open the file for the sequences with address  " << seqFile <<endl; exit(1); }
  
  if  ((cFpWriteFB = fopen("FBprob.txt", "w" ))  == NULL )
    { cerr << "\n could not open the file FBprob for writing  \n"<<endl; exit(1); }

  if (FBDdetailsC)
    {
      if  ( (cFpWriteCP = fopen("PMatrix.txt","w" ))  == NULL )
	{ cerr << "\n could not open the file PMatrix for writing \n";}
    }
  completedFile = false; seqNb = 0;
while (! completedFile)
  {
      seqNb++;seqLen = 0;
      if ( ! (ReadSeq (seqNb, seqLen, seq, completedFile, SeqName) )   )
	  {
	  if (verbose)  cerr << "\nReadSeq done" <<  endl;
	  if (seqLen > 0) { WriteWarning(seqNb,1);}	
	  else {seqNb--;}
	  } 
      if (seqLen > 0)
	  { 
	  tok = strtok(SeqName, "\t  \n");
	  fprintf(cFpWriteFB, "%s",tok);
	  if (verbose)   cerr << ">>read seq-Nb " << seqNb << "  name = " << SeqName << endl;
	  if (verbose)   cerr << "processing seq-Nb " << seqNb << "  name = " << SeqName << endl;

	  if (bUseExtP)
	  	{
         if (seqLen > kMaxExtPLen)
{cerr<<"\n Sequence Length excessive for use of external prob matrix, limit= "<<kMaxExtPLen<<endl;}
		 else 
		    {
			model.FBalgorithmExtP(seqNb, seqLen, seq);
	    	fprintf(cFpWriteFB, " #ln(FB-P)=%6.15f\n", (double) model.logFBseqProb );
	    	if (FBDdetailsC) {WriteProbProfile (seqLen,seq);}
			}
		completedFile = true;// only one sequence with bUseExtP
		}
	  else 
	  	{
		model.FBalgorithm(seqNb, seqLen, seq);
	    fprintf(cFpWriteFB, " #ln(FB-P)=%6.15f\n", (double) model.logFBseqProb );
	    if (FBDdetailsC) {WriteProbProfile (seqLen,seq);}
	    if (bdsDNA) 
	    	{
	      model.rcseq(seqLen, seq);
	      seqNb +=1;
	      fprintf(cFpWriteFB, "%sSecondStrand",tok);
	      model.FBalgorithm(seqNb, seqLen, model.SecondStrand);
	      fprintf(cFpWriteFB, " #ln(FB-P)SecondStrand=%6.15f\n", (double) model.logFBseqProb );
	      if (FBDdetailsC) {WriteProbProfile (seqLen, model.SecondStrand);}
	    }	}
	  }
      if ( (seqNb % kperiodStdout1) == 0)	 
	  {
	  if (seqNb == kperiodStdout1) cerr<<"One point "<<kperiodStdout1<<" sequences, one line "<<kperiodStdout2<<"\n"; 
	  cerr<<".";  //fflush (stdout);
	  }
      if ( (seqNb % kperiodStdout2) == 0)  cerr<<"\n";  	 
      if ( (seqNb % kperiodStdout3) == 0)  cerr<<"\n processed seq-nb= "<<seqNb<<"\n";
  }
  
  fclose(cFpReadSeq);
  fprintf (cFpWriteFB,"\n");fclose(cFpWriteFB);
  if (FBDdetailsC) {fprintf (cFpWriteCP,"\n"); fclose( cFpWriteCP);}
  
  return;
}

//---------------------------------------------------------------------------
//		 WriteProbProfile 		WRITE MATRIX OUTPUT			
//---------------------------------------------------------------------------
void  WriteProbProfile(int len, const int Seq[])
{
int i, pos, st, nbStates;  

nbStates = model.nbStates;
i=0;

if (seqNb==1) 
{
fprintf (cFpWriteCP, "Nsequence\tNposition\tSEQ");

for (i=0; i<nbStates; i++)
	{
	fprintf (cFpWriteCP, "\t%s", model.statenames[i]);	
	}
}

pos=0;	st=0;
fprintf (cFpWriteCP,"\n%d\t%d\tXX\t%6.3f", seqNb, pos, (double) model.PosStateProb[pos][st]);

for (st=1; st<nbStates; st++)
	{
	fprintf (cFpWriteCP, "\t%6.3f", (double) model.PosStateProb[pos][st]);
	}

for (pos=1; pos<(len+1); pos++)
	{
	st=0;
	fprintf (cFpWriteCP, "\n%d\t%d\t%c\t%6.3f", seqNb, pos, (char) Decode[Seq[pos]], (double) model.PosStateProb[pos][st]);

	for (st=1; st<nbStates; st++)
		{
		fprintf (cFpWriteCP, "\t%6.3f", (double) model.PosStateProb[pos][st]);
		}
	}

pos=(len+1);st=0;
fprintf (cFpWriteCP, "\n%d\t%d\tXX\t%6.3f", seqNb, pos, (double) model.PosStateProb[pos][st]);
for (st=1; st<nbStates; st++)
	{
	fprintf (cFpWriteCP, "\t%6.3f", (double) model.PosStateProb[pos][st]);
	}

}
/*  WriteProbProfile
************************************/
