/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

extern class ModelFB model;
extern int vMAXnbITERATIONS;
extern int vMINnbITERATIONS;
extern int maxlen;
extern float vMINdifftotLogLik;
extern bool alloutput, storeseqs;

#include "VitLearnMain.h"

extern bool verbose, optCmatrix, bUseExtP;	
extern  int  Code[], Decode[];
extern bool bWriteSeq;
extern double thresholds[6];
extern int    thresholdNb;
extern  int   seqNb, seqLen;
extern  int   seq[kMaxSeqLen+1];
extern  bool  completedFile;
extern  char  SeqName[kMaxSeqName+2];
extern  FILE *cFpWriteDom; 
extern  FILE *cFpWritePP, *cFpWritePPD,*cFpWriteCP; 
extern  FILE *cFpReadSeq;
extern bool   bdsDNA, bdsDNAjointly;

// defined in BW
extern  int iterationNb;
extern  long double totLogLik, origtotLogLik, prevtotLogLik, difftotLogLik;
extern  FILE *cFpWriteBW;
// defined in Viterbi decoding
//extern  FILE *cFpWriteVITDEC;

bool CheckStopVL(void);
void MonitorLikChangeVL(void);

// ---------------------------------------------------------------------------
//		 VitLearnMain 				
// ---------------------------------------------------------------------------
void VitLearnMain(const char *ModelFile, char *seqFile, const int nbseqs, const int maxlen )
{
bool stoppingCondition;
int i, j, totseqnb=0;
int **seqs=NULL; 
int *seqsmem;
int *seqsLen=NULL;

model.doingVitLearning = true;
	
if (verbose) cerr << "VitLearnMain" << endl;

if (storeseqs)// keep seqs in memory
	{
	if ((nbseqs>0) && (maxlen>0))
		{
		// create storage space
		seqsmem=(int *) malloc(nbseqs*(maxlen+1)*sizeof(int));
		if (seqsmem==NULL) {cerr<<"\nFailed in malloc seqsmem in VitLearnMain"<<endl;exit(0);}
		seqs=(int **) malloc(nbseqs*sizeof(int *));
		if (seqs==NULL) {cerr<<"\nFailed in malloc seqsin VitLearnMain"<<endl;exit(0);}
		for (i=0; i<nbseqs; i++) seqs[i]=seqsmem+i*(maxlen+1);
		for (i=0; i<nbseqs; i++)
			{
			for (j=0; j<=maxlen; j++)  seqs[i][j]=0;
			}

		seqsLen=(int *) malloc(nbseqs*sizeof(int));
		if (seqsmem==NULL) {cerr<<"\nFailed in malloc seqsLen in VitLearnMain"<<endl;exit(0);}
		}
	else {cerr<<"\nThe nb of sequences and their maximal length are required with option -k"<<endl;exit(0);}
	}

iterationNb = 0; stoppingCondition = false;

if  ((cFpWriteBW = fopen( "InitialModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file InitialModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#BEFORE VITERBI-LEARNING, INITIAL MODEL \n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

if (alloutput)
{
if  ((cFpWriteBW = fopen( "UpdateModels.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file UpdateModels for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"\n1.BEFORE VITERBI-LEARNING, INITIAL MODEL\n");
model.WriteModelFB(cFpWriteBW);
}


if  ((cFpReadSeq = fopen( seqFile, "r" ))  == NULL )
		{cerr << "\n could not open the file for the sequences with address  " << 
					seqFile <<endl;  exit(1);}

while (stoppingCondition == false )
	{
	model.SetUpLogParameters();  
	model.SetUpConnections();
	model.InitialiseBWIterStep();
		  	
	seqNb = 0;totLogLik=0;
	completedFile = false;iterationNb++;

	if ((storeseqs) && (iterationNb>1))  
		{
		while (seqNb < totseqnb)
			{
			seqNb ++;
			seqLen = seqsLen[seqNb];
			if (seqLen > 0)
				{ 
				  if (bUseExtP)
	  				{
        			 if (seqLen > kMaxExtPLen)
			{cerr<<"\n Sequence Length excessive for use of external prob matrix, limit= "<<kMaxExtPLen<<endl;}
					 else 
		    			{
						model.ViterbiLextP(seqNb, seqLen, seqs[seqNb]);
						totLogLik+=  model.ViterbiPathLogProb;
						model.ComputeSeqVL(seqNb, seqLen, seqs[seqNb]); 
						}
					seqNb = totseqnb;// only one sequence with bUseExtP
					}
				  else 
					{
					model.ViterbiL(seqNb, seqLen, seqs[seqNb]);
					totLogLik+=  model.ViterbiPathLogProb;
					model.ComputeSeqVL(seqNb, seqLen, seqs[seqNb]); 
					}
				}	
			}
		}
		
	else
		{
		rewind(cFpReadSeq);
		while (completedFile == false)
			{
			seqNb ++;
			if ( !(ReadSeq (seqNb, seqLen, seq, completedFile, SeqName)) )
				{
				if (seqLen > 0)	{ WriteWarning(seqNb,1);}	
				else	{seqNb--;}
				} 
			if (seqLen > 0)
				{
				  if (bUseExtP)
	  				{
        			 if (seqLen > kMaxExtPLen)
			{cerr<<"\n Sequence Length excessive for use of external prob matrix, limit= "<<kMaxExtPLen<<endl;}
					 else 
		    			{
						model.ViterbiLextP(seqNb, seqLen, seq);
						totLogLik+=  model.ViterbiPathLogProb;
						model.ComputeSeqVL(seqNb, seqLen, seq); 
						}
					seqNb = totseqnb;// only one sequence with bUseExtP
					}
				  else 
					{
					model.ViterbiL(seqNb, seqLen, seq);
					totLogLik+=  model.ViterbiPathLogProb;
					model.ComputeSeqVL(seqNb, seqLen, seq); 
					}
				}	
			if ((storeseqs) && (iterationNb==1))
				{
				for (j=1; j<=seqLen; j++)  seqs[seqNb][j]=seq[j];
				seqsLen[seqNb] = seqLen;
 				}
			}
		totseqnb = 	seqNb;

		}
	if (iterationNb==1) {prevtotLogLik=origtotLogLik=totLogLik;}
	model.ParameterIterationBW(); 
	if (alloutput) 
		{
		fprintf( cFpWriteBW,"\n#AFTER ITERATION NB=%d\n",iterationNb);
		model.WriteModelFB(cFpWriteBW);
		}
	MonitorLikChangeVL();
	stoppingCondition = CheckStopVL();
	}
	
if (alloutput)
{ 
fprintf(cFpWriteBW,"\n#AFTER LAST ITERATION\n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);
}

if  ((cFpWriteBW = fopen( "FinalModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file FinalModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#AFTER VITERBI LEARNING, FINAL MODEL\n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

//fclose(cFpWriteVITDEC);

fclose(cFpReadSeq);
}
/* VITERBI-LEARNING
************************************/



// ---------------------------------------------------------------------------
//		 VitLearnMaindsDNA 				
// ---------------------------------------------------------------------------
void VitLearnMaindsDNA(const char *ModelFile, char *seqFile, const int nbseqs, const int maxlen )
{
bool stoppingCondition;
int i, j, totseqnb=0;
int **seqs=NULL; 
int *seqsmem;
int *seqsLen=NULL;
long double firststrandloglik, secondstrandloglik, diff, secondstrandweight, P=0;

model.doingVitLearning = true;
	
if (verbose) cerr << "VitLearnMain" << endl;

if (storeseqs)// keep seqs in memory
	{
	if ((nbseqs>0) && (maxlen>0))
		{
		seqsmem=(int *) malloc(nbseqs*(maxlen+1)*sizeof(int));
		if (seqsmem==NULL) {cerr<<"\nFailed in malloc seqsmem in VitLearnMain"<<endl;exit(0);}
		seqs=(int **) malloc(nbseqs*sizeof(int *));
		if (seqs==NULL) {cerr<<"\nFailed in malloc seqsin VitLearnMain"<<endl;exit(0);}
		for (i=0; i<nbseqs; i++) seqs[i]=seqsmem+i*(maxlen+1);
		for (i=0; i<nbseqs; i++)
			{
			for (j=0; j<=maxlen; j++)  seqs[i][j]=0;
			}

		seqsLen=(int *) malloc(nbseqs*sizeof(int));
		if (seqsmem==NULL) {cerr<<"\nFailed in malloc seqsLen in VitLearnMain"<<endl;exit(0);}
		}
	else {cerr<<"\nThe nb of sequences and their maximal length are required with option -k"<<endl;exit(0);}
	}

iterationNb = 0;stoppingCondition = false;

if  ((cFpWriteBW = fopen( "InitialModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file InitialModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#BEFORE VITERBI-LEARNING, INITIAL MODEL \n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

if (alloutput)
{
if  ((cFpWriteBW = fopen( "UpdateModels.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file UpdateModels for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"\n1.BEFORE VITERBI-LEARNING, INITIAL MODEL\n");
model.WriteModelFB(cFpWriteBW);
}

if  ((cFpReadSeq = fopen( seqFile, "r" ))  == NULL )
		{cerr << "\n could not open the file for the sequences with address  " << 
					seqFile <<endl;  exit(1);}

while (stoppingCondition == false )
	{
	model.SetUpLogParameters();  
	model.SetUpConnections();
	model.InitialiseBWIterStep();
		  	
	seqNb = 0;totLogLik=0;
	completedFile = false;iterationNb++;

	if ((storeseqs) && (iterationNb>1))  
		{
		while (seqNb < totseqnb)
			{
			seqNb ++;
			seqLen = seqsLen[seqNb];
			if (seqLen > 0)
				{ 
				model.ViterbiL(seqNb, seqLen, seq);
				totLogLik+=  model.ViterbiPathLogProb;
				model.ComputeSeqVL(seqNb, seqLen, seq); 
				}	
			}
		}
		
	else
		{
		rewind(cFpReadSeq);
		while (completedFile == false)
			{
			seqNb ++;
			if ( !(ReadSeq (seqNb, seqLen, seq, completedFile, SeqName)) )
				{
				if (seqLen > 0)	{ WriteWarning(seqNb,1);}	
				else	{seqNb--;}
				} 
			if (seqLen > 0)
				{
				model.ViterbiL(seqNb, seqLen, seq);

				if (bdsDNA) {totLogLik+= model.ViterbiPathLogProb;} //as indep. sequences (not my approach) 
				else {P = exp(model.ViterbiPathLogProb);}
				firststrandloglik = model.ViterbiPathLogProb; 
				model.ComputeSeqVLdsDNA1(seqNb, seqLen, seq);
				model.rcseq(seqLen, seq);
				model.ViterbiL(seqNb, seqLen, model.SecondStrand);
				secondstrandloglik = model.logFBseqProb;
				diff = firststrandloglik - secondstrandloglik; 
				if  (bdsDNAjointly) {secondstrandweight = 1.0 / (1.0 + ((long double) exp ( (double) diff)));}
				else {secondstrandweight = 0.5 ; }
				model.ComputeSeqVLdsDNA2(seqNb, seqLen, model.SecondStrand, secondstrandweight);
				if (bdsDNA) {totLogLik+= model.ViterbiPathLogProb;} 
				else {totLogLik+= log(P+exp(model.logFBseqProb) );}

				
				}	
			if ((storeseqs) && (iterationNb==1))
				{
				for (j=1; j<=seqLen; j++)  seqs[seqNb][j]=seq[j];
				seqsLen[seqNb] = seqLen;
 				}
			}
		totseqnb = 	seqNb;
		}
		
	if (iterationNb==1) {prevtotLogLik=origtotLogLik=totLogLik;}
	model.ParameterIterationBW();  
	if (alloutput) 
		{
		fprintf( cFpWriteBW,"\n#AFTER ITERATION NB=%d\n",iterationNb);
		model.WriteModelFB(cFpWriteBW);
		}
	MonitorLikChangeVL();
	stoppingCondition = CheckStopVL();
	}
	
if (alloutput)
{ 
fprintf(cFpWriteBW,"\n#AFTER LAST ITERATION\n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);
}

if  ((cFpWriteBW = fopen( "FinalModel.txt", "w" ))  == NULL )
		{cerr << "\n could not open the file FinalModel for writing"<<endl; exit(1);}
fprintf( cFpWriteBW,"#AFTER VITERBI LEARNING, FINAL MODEL\n");
model.WriteModelFB(cFpWriteBW);
fclose(cFpWriteBW);

//fclose(cFpWriteVITDEC);

fclose(cFpReadSeq);
}
/* VITERBI-LEARNING dsDNA
************************************/

// ---------------------------------------------------------------------------
//		CheckStopVL				
// ---------------------------------------------------------------------------
bool   CheckStopVL(void)  
{
bool  decision;

decision = false;
if (iterationNb >= vMAXnbITERATIONS)  decision = true;
if ((iterationNb >= vMINnbITERATIONS) && ( fabs(difftotLogLik) < vMINdifftotLogLik))  decision = true;
if (iterationNb < 2) decision = false;	

return(decision);
}
/* CheckStopVL
************************************/

// ---------------------------------------------------------------------------
//		 MonitorLikChangeVL	> 				
// ---------------------------------------------------------------------------
void MonitorLikChangeVL(void)
{
difftotLogLik=totLogLik-prevtotLogLik;
cerr<<"\n logLik of the learning set of sequences   " << totLogLik << "\n at evaluation " <<  iterationNb;
cout<<"\n logLik of sequence set=  " << totLogLik;
cout<<"\n difftotLogLik of sequence set=  " << difftotLogLik;
cout<<"\n prevtotLogLik of sequence set=  " << totLogLik;
cout<<"\n origtotLogLik of sequence set=  " << origtotLogLik;
cout<<"\n" << endl;

if (alloutput) 
	{fprintf(cFpWriteBW,  "\nORIG logLik=%g \nPREV logLik=%g \t PRES logLik=%g \t DIFF logLik=%g \t   "  ,
		(double) origtotLogLik, (double) prevtotLogLik,(double) totLogLik, (double) difftotLogLik);
	} 
prevtotLogLik=totLogLik;
}
/* MonitorLikChangeVL
************************************/
////////////////////////////////////////////////////////////////////////////////////////////////////
