/***********************************************
************************************************
**                                            **
** mamot                                      **
** a program for HMM modelling                **
** 2008, Mauro C. Delorenzi                   **
**                                            **
************************************************
************************************************/

//prevent multiple inclusion
#ifndef ModelFB_H
#define ModelFB_H

#include <cstdlib> 
#include <stdio.h>
#include <stdlib.h> 

#include <iostream.h> 
#include <fstream.h> 
#include <iomanip.h>  

#include <math.h>
#include <float.h>
#include <string.h>

#include "globals.h"

class ModelFB
	{
	public: 
//vars
	int MaxNameLen;
	int startstate, endstate, nbEMties, nbcEMties;
	int usemodus;
	long double  logFBseqProb;
	bool possible, bFBseqProbPositive;
 	bool doingVitLearning;
	char letter[kMaxNbLetters];   //alphabet *letter;  
	int SecondStrand[kLength+1];
	int nbLetters, nbStates, maxSeqLength;
	double **PosStateProb; double *PosStateProbmem;
	char **statenames; char *statenamesmem;
	long double ViterbiPathLogProb;

//methods
	ModelFB ();      
	~ModelFB ();	
	void  AssignMemory (int modus, int s, int a, int l, int m, int t, int T, int c);
	//constructor, initialise dynamic memory

	void SetUpScaledParameters(void);
	void SetUpScaledParametersW(void);
	void SetUpLogParameters(void);
	void SetUpLogParametersW(void);

	void SetUpConnections(void);
	bool ReadModelFB (FILE *cFrMod,FILE *cFpWriteModel );
	bool ReadextP (FILE *cFrMod);  


	void InitCodes(void);
	void WriteModelFB (FILE *cFpWriteModel);
		
	void Viterbi (int seqNb, int totalNumber, const int Seq[], FILE *cFpWriteVITDEC); 	
	void ViterbiExtP (int seqNb, int totalNumber, const int Seq[], FILE *cFpWriteVITDEC); 	
	void ViterbiL (int seqNb, int totalNumber, const int Seq[]); 	
	void ViterbiLextP (int seqNb, int totalNumber, const int Seq[]); 	

	void FBalgorithm (int seqNb, int totalNumber, const int Seq[]); 
	void FBalgorithmExtP (int seqNb, int totalNumber, const int Seq[]); 
	
	void GenerateSequence (FILE *cFpWriteGenSeq, int genseqnb, int maxlen);
    void GenerateSequenceB (int genseqnb, int maxlen);

	void ParameterIterationBW(void);
	void Pseudocounts( void );

	void ComputeSeqBW(int seqNb, int totalNumber, const int Seq[]);
	void ComputeSeqBWdsDNA1(int seqNb, int totalNumber, const int Seq[]);
	void ComputeSeqBWdsDNA2(int seqNb, int totalNumber, const int Seq[], long double secstrandweight);

	void ComputeSeqVL(int seqNb, int totalNumber, const int Seq[]);
	void ComputeSeqVLdsDNA1(int seqNb, int totalNumber, const int Seq[]);
	void ComputeSeqVLdsDNA2(int seqNb, int totalNumber, const int Seq[], long double secstrandweight);

	void InitialiseBWIterStep(void);
	
	int  RndEmission(const int s); 
	int  RndStateTransition(const int s); 
	int  bestPathM[kMaxSeqLen];
	float ran2(long &idum );
    void Initcum(); 
	
	void rcseq(int totalNumber, int Seq[]); 

	private: 
//vars
   // dynamically assigned memory, once only in the constructor 
	int **ties; int *tiesmem;
	int **cties; int *ctiesmem;
	long double **emProb; long double *emProbmem;
	long double **trProb; long double *trProbmem;
	long double **scEmProb; long double *scEmProbmem;
	double **extP;  double *extPmem;
	long double **scTrProb; long double *scTrProbmem;
	long double **scForw; long double *scForwmem;
	int **tracebackM; int *tracebackMmem;
	long double **scBackw; long double *scBackwmem;
	long double emScale;
	//for BW
	long double **Cont; long double *Contmem; 
	long double **tCont; long double *tContmem;
	long double **Cont2; long double *Contmem2; 
	long double **tCont2; long double *tContmem2;
	long double **scSumEmTot; long double *scSumEmTotmem;
	long double **scSumTrTot; long double *scSumTrTotmem;

	double *pseudocountsEm, *nullEmFreq, *weightEM;
	int *ForwConnList, *BackConnList;
	int *ForwConnFirst, *BackConnFirst;
	int *ForwConnLast, *BackConnLast;
	int lastSeqBW;
	int FBFactor[kMaxSeqLen];
	long double FBscaleFactor;
	int *updatetr, *updateem, *pseudow;  
	};
#endif
